#!/usr/bin/env python3

import atexit
import logging
import redis, threading, traceback, sys

from configator.engine import RedisClient
from typing import Any, Callable, List, Tuple, Dict, Optional

LOG = logging.getLogger(__name__)

class SettingSubscriber(RedisClient):
    #
    def __init__(self, *args, auto_stop=True, **kwargs):
        if auto_stop:
            atexit.register(self.stop)
        super(SettingSubscriber, self).__init__(**kwargs)
        self.CHANNEL_PATTERN = self.CHANNEL_GROUP + '*'
    #
    ##
    @property
    def pubsub(self):
        ps = self.connect().pubsub()
        ps.psubscribe(**{self.CHANNEL_PATTERN: self.__process_event})
        return ps
    #
    ##
    __t = None
    #
    def start(self):
        if self.__t is None:
            self.__t = self.__run_in_thread(sleep_time=0.001)
            if LOG.isEnabledFor(logging.DEBUG):
                LOG.log(logging.DEBUG, "SettingSubscriber has started")
        return self.__t
    #
    def stop(self):
        if self.__t is not None:
            self.__t.stop()
            self.__t = None
        self._destroy()
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.log(logging.DEBUG, "SettingSubscriber has stopped")
    #
    def __run_in_thread(self, sleep_time=0, daemon=False):
        thread = PubSubWorkerThread(self, sleep_time, daemon=daemon)
        thread.start()
        return thread
    #
    ##
    __transformer: Optional[Callable[[Dict], Tuple[Dict, Any]]] = None
    #
    def set_transformer(self, transformer: Callable[[Dict], Tuple[Dict, Any]]):
        if callable(transformer):
            self.__transformer = transformer
        return self
    #
    ##
    __event_mappings: Optional[Dict[Callable[[Dict, Any], bool], Tuple[Callable[[Dict, Any], None],...]]] = None
    #
    def add_event_handler(self, match: Callable[[Dict, Any], bool], *reset: Callable[[Dict, Any], None]):
        if self.__event_mappings is None:
            self.__event_mappings = dict()
        if not callable(match):
            raise ValueError('match must be callable')
        if not reset:
            raise ValueError('reset list must not be empty')
        self.__event_mappings[match] = reset
        return self
    #
    def __process_event(self, message):
        if self.__event_mappings is None:
            return
        #
        if self.__transformer is not None:
            msg, err = self.__transformer(message)
        else:
            msg, err = (message, None)
        #
        for match, reaction in self.__event_mappings.items():
            if match(msg, err):
                for reset in reaction:
                    if callable(reset):
                        reset(msg, err)


class PubSubWorkerThread(threading.Thread):
    def __init__(self, redis_client, sleep_time, daemon=False):
        super(PubSubWorkerThread, self).__init__()
        self.daemon = daemon
        #
        self._redis_client = redis_client
        self._sleep_time = sleep_time
        #
        self._running = threading.Event()
    #
    #
    def run(self):
        if self._running.is_set():
            return
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.log(logging.DEBUG, "PubSubWorkerThread is starting")
        self._running.set()
        while self._running.is_set():
            try:
                pubsub = self._redis_client.pubsub
                while self._running.is_set():
                    pubsub.get_message(ignore_subscribe_messages=True, timeout=self._sleep_time)
                pubsub.close()
            except redis.ConnectionError:
                self._redis_client.reconnect()
            except Exception as err:
                traceback.print_exc(file=sys.stdout)
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.log(logging.DEBUG, "PubSubWorkerThread has stopped")
    #
    #
    def stop(self):
        self._running.clear()



#!/usr/bin/env python3

import json
from typing import Any, Callable, List, Tuple, Dict, Optional


def json_loads(data):
    if not isinstance(data, str):
        return data, None
    try:
        json_dict = json.loads(data)
        return json_dict, None
    except Exception as exception:
        return data, exception


def transform_json_data(message:Dict) -> Tuple[Dict, Any]:
    if not isinstance(message, dict):
        return message, ValueError('message must be a dict')
    #
    if 'data' not in message:
        return message, ValueError('data field not found')
    #
    if not isinstance(message['data'], bytes):
        return message, ValueError('message[data] must be bytes or string')
    #
    obj, err = json_loads(message['data'].decode('utf-8'))
    if obj:
        message['data'] = obj
    return message, err


def match_by_label(name):
    assert isinstance(name, bytes)
    def match_func(message, err, *args, **kwargs):
        if not isinstance(message, dict):
            return False
        #
        channel_name = message.get('channel')
        if not isinstance(channel_name, bytes):
            return False
        #
        return channel_name.endswith(name)
    return match_func



#!/usr/bin/env python3

import logging
import json

from configator.engine import RedisClient
from typing import List, Tuple, Dict, Optional, Union

LOG = logging.getLogger(__name__)

class SettingPublisher(RedisClient):
    #
    def __init__(self, *args, **kwargs):
        super(SettingPublisher, self).__init__(**kwargs)
        self.CHANNEL_PREFIX = self.CHANNEL_GROUP + ':'
    #
    #
    def publish(self, message: Union[Dict, bytes, str, int, float], label:Optional[str]=None):
        try:
            self.publish_or_error(message, label=label)
            return None
        except Exception as err:
            return err
    #
    #
    def publish_or_error(self, message: Union[Dict, bytes, str, int, float], label:Optional[str]=None):
        if label is None:
            channel_name = self.CHANNEL_GROUP
        else:
            channel_name = self.CHANNEL_PREFIX + label
        #
        if isinstance(message, dict):
            message = json.dumps(message)
        elif not self.__is_valid_type(message):
            errmsg = "Invalid type of input: '%s'. Only a dict, bytes, string, int or float accepted." % type(message)
            if LOG.isEnabledFor(logging.ERROR):
                LOG.log(logging.ERROR, errmsg)
            raise ValueError(errmsg)
        #
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.log(logging.DEBUG, "publish() a message [%s] to channel [%s]", str(message), channel_name)
        #
        self.connect().publish(channel_name, message)
    #
    #
    @staticmethod
    def __is_valid_type(data):
        return isinstance(data, (bytes, str, float)) or (isinstance(data, int) and (type(data) != type(True)))



#!/usr/bin/env python3

import logging
import redis, os

from configator.utils.datatype import str_to_int

DEFAULT_CHANNEL_GROUP = 'configator'
DEFAULT_ENV_PREFIX = 'CONFIGATOR'

LOG = logging.getLogger(__name__)

class RedisClient(object):
    #
    def __init__(self, channel_group=None, env_prefix=None, **connection_kwargs):
        #
        self.CHANNEL_GROUP = channel_group if isinstance(channel_group, str) else DEFAULT_CHANNEL_GROUP
        #
        self.ENV_PREFIX = env_prefix if isinstance(env_prefix, str) else DEFAULT_ENV_PREFIX
        env_prefix_lodash = self.ENV_PREFIX + '_'
        #
        self.__connection_kwargs = connection_kwargs
        #
        host = os.getenv(env_prefix_lodash + 'REDIS_HOST')
        if host:
            self.__connection_kwargs['host'] = host
        #
        if not self.__connection_kwargs.get('host'):
            self.__connection_kwargs['host'] = 'localhost'
        #
        port = os.getenv(env_prefix_lodash + 'REDIS_PORT')
        if port:
            port, err = str_to_int(port)
            if err is None and port > 0:
                self.__connection_kwargs['port'] = port
        if not self.__connection_kwargs.get('port'):
            self.__connection_kwargs['port'] = 6379
        #
        db = os.getenv(env_prefix_lodash + 'REDIS_DB')
        if db:
            db, err = str_to_int(db)
            if err is None and type(db) == type(0):
                self.__connection_kwargs['db'] = db
        if 'db' not in self.__connection_kwargs:
            self.__connection_kwargs['db'] = 0
        #
        username = os.getenv(env_prefix_lodash + 'REDIS_USERNAME')
        if username:
            self.__connection_kwargs['username'] = username
        #
        password = os.getenv(env_prefix_lodash + 'REDIS_PASSWORD')
        if password:
            self.__connection_kwargs['password'] = password
        #
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.log(logging.DEBUG, "redis connection kwargs: %s", str(self.__connection_kwargs))
    #
    ##
    def connect(self, ping:bool=False):
        waiting = True
        while waiting:
            try:
                connection = self._connection
                if ping:
                    connection.ping()
                waiting = False
                return connection
            except redis.ConnectionError:
                if LOG.isEnabledFor(logging.DEBUG):
                    LOG.log(logging.DEBUG, "redis.ConnectionError")
                pass
    #
    #
    def reconnect(self):
        self._destroy()
        return self.connect()
    #
    #
    __r = None
    #
    #
    @property
    def _connection(self):
        if self.__r is None:
            pool = redis.ConnectionPool(**self.__connection_kwargs)
            self.__r = redis.Redis(connection_pool=pool)
        return self.__r
    #
    #
    def _destroy(self):
        if self.__r is not None:
            self.__r.close()
            self.__r = None



#!/usr/bin/env python

import setuptools

setuptools.setup(
  name = 'configator',
  version = '0.1.0',
  description = 'A simple configuration loader',
  author = 'acegik',
  license = 'GPL-3.0',
  url = 'https://github.com/acegik/configator',
  download_url = 'https://github.com/acegik/configator/downloads',
  keywords = ['configuration'],
  classifiers = [],
  install_requires = open("requirements.txt").readlines(),
  python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*",
  package_dir = {'':'src'},
  packages = setuptools.find_packages('src'),
)
